/*
 * chk.c
 *
 * Check the grammar for errors, ambiguities etc...
 *
 * Forced variable names to be unique to 7 char for portability
 * Terence Parr -- March 1990
 */

#include <stdio.h>
#ifdef MEMCHK
#include "trax.h"
#endif
#include "set.h"
#include "struct.h"
#include "attrib.h"

/*
 * Compute firsts of all rules
 */
CheckRules(r)
Atom ***r;
{
	set *f,t;
	int i;
	
	for (i=1; r[i] != NULL; i++)
	{
		chkBlk(r[i], i);
		f = First(i);
		if ( PrintFi ) {
			printf("First(%s) is\n", RuleStr[i]);
			fset_pr( f );
		}
		ambigRule(f, i);
		freeFset(f);
		
		if ( PrintFo ) {
			t = Follow(i);
			printf("Follow(%s) is ", RuleStr[i]);
			set_prT(t);
			printf("\n\n");
			set_free(t);
		}
	}
}

/*
 * chkBblk()
 *
 * Check sub block for problems.
 */
static
chkBlk(b, rule)
Atom **b;
int rule;
{
	int alt, atm;

	for (alt=1; b[alt] != NULL; alt++)
	{
		for (atm=1; b[alt][atm].type != EOL; atm++)
		{
			chkAtom( &(b[alt][atm]), rule );
		}
	}
}

static
chkAtom( a, rule )
Atom *a;
int rule;
{
	set *f;

	switch ( a->type )
	{
		case aRule :
			break;
		case aSubBLK :
		case aOptBLK :
		case aLoopBLK :
			if ( a->ptr.block == NULL ) break;
			if ( emptyBlk(a->ptr.block) ) break;
			fAtom(a, rule, FALSE, &f);
			ambigSub(f, a, rule);
			if ( PrFirstSub && PrintFi )
			{
				printf("First[");
				dispAtom( *a, stdout );
				printf(" ] is\n");
				fset_pr( f );
			}
			chkBlk(a->ptr.block, rule);
			freeFset( f );
			break;
	}
}

ambiguous( f, t, a1, a2 )
set *f;
int *t, *a1, *a2;
{
	int alt, other;
	set g;

	*a1 = *a2 =0;
	for (alt = 0; f[alt].n != 0; alt++)	/* for each alt */
	{
		*a1 = alt;
		for (other=alt+1; f[other].n != 0; other++)
		{
			*a2 = other;
			g = set_and(f[alt], f[other]);
			if ( set_deg(g) > 0 )
			{
				*t = set_int(g);		/* Return 1st Ambiguity */
				set_free(g);
				return( TRUE );
			}
			set_free(g);
		}
	}
	return( FALSE );
}

ambigRule(f,r)
set *f;
int r;
{
	unsigned t, a1, a2, alt;

	if ( ambiguous( f, &t, &a1, &a2 ) )
	{
		if ( a2==0 )
		{
			warnFL( eMsg("warning: alt %d rule %s ambiguous upon %s",
					a1+1,RuleStr[r],TokenName(t)),
				FileStr[RuleFile[r]], RuleLine[r] );
		}
		else
		{
			warnFL( eMsg("warning: alts %d,%d rule %s ambiguous upon %s",
					a1+1,a2+1,RuleStr[r],TokenName(t)),
				FileStr[RuleFile[r]], RuleLine[r] );
		}
	}
	for (alt=0; f[alt].n != 0; alt++)
	{
		if ( set_el(EpToken, f[alt]) )
		{
			warnFL( eMsg("warning: alt %d rule %s empty follow set; assuming EOF",
						alt+1, RuleStr[r]),
					FileStr[RuleFile[r]], RuleLine[r] );
			set_rm(EpToken, f[alt]);
			set_orel(EofToken, &(f[alt]));
		}
	}
}

ambigSub(f, a, r)
set *f;
Atom *a;
unsigned r;
{
	unsigned t,a1,a2,alt;

	if ( ambiguous( f, &t, &a1, &a2 ) )
	{
		warnBlk( *a, eMsg(" [rule %s] ambiguous upon %s",
				RuleStr[r], TokenName(t)) );
	}
	for (alt=0; f[alt].n != 0; alt++)
	{
		if ( set_el(EpToken, f[alt]) )
		{
			warnBlk( *a, eMsg(" [rule %s] empty follow set; assuming EOF",
							RuleStr[r]) );
			set_rm(EpToken, f[alt]);
			set_orel(EofToken, &(f[alt]));
		}
	}
}

InhChk(rule)
Atom ***rule;
{
	register int r;
	
	for (r=1; rule[r] != NULL; r++)
	{
		InhChkBlk(rule[r], r);
	}
}

InhChkBlk(blk, r)
Atom **blk;
int r;
{
	register Atom *alt;
	register int a, atm;
	int t;
	
	for (a=1; blk[a] != NULL; a++)
	{
		alt = blk[a];
		for (atm=1; alt[atm].type != EOL; atm++)
		{
			switch ( alt[atm].type )
			{
				case aSubBLK :
				case aOptBLK :
				case aLoopBLK :
					InhChkBlk(alt[atm].ptr.block, r);
					break;
				case aRule :
					if ( (t=ruleNum(&(alt[atm]))) == 0 ) continue;
					/* Do rule cross reference */
					/* Check Inheritance override consistency */
					if ( alt[atm].inheritance != NULL )
					{
						if ( !set_el(t, RuleInh) && !set_nil( CrossRef[t]) )
						{
							warnFL(
								eMsg("warning: inheritance for rule %s used inconsistently in rule %s",
										 alt[atm].ptr.rule,
										 RuleStr[r]),
								FileStr[RuleFile[r]], RuleLine[r] );
						}
						set_orel(t, &RuleInh);
					}
					else
					{
						if ( set_el(t, RuleInh) )
						{
							warnFL(
								eMsg("warning: inheritance for rule %s used inconsistently in rule %s",
										 alt[atm].ptr.rule,
										 RuleStr[r]),
								FileStr[RuleFile[r]], RuleLine[r] );
						}
					}
					set_orel(r, &(CrossRef[t]));
					break;
			}
		}
	}
}
