/* parser that follows ALX grammar for ee595e
 *
 * Will Cohen
 * 10/3/89
 *
 *
 * the lexical analyser converts the '%%', '|', '[', ']', '(', ')',
 * '{', '}', '+', '*', '-', '!' into tokens which are larger than MAXCHAR.
 * This is done so there is less confusion on what a character is when
 * it is time to build the nfa.
 *

* more than one action allowed at begin and end since ALX allows it *
grammar		: (action)* '%%' rulelist '%%' (action)*
		;

rule_list	: rule (rule)*
		;

rule		: regexpr action
		;

regexpr		: and_expr ('|' andexpr)*
		;

andexpr		: expr (expr)*
		;

expr		: {'~'} '[' atomlist ']' {repeatsymbol}
		| '(' regexpr ')' {repeatsymbol}
		| '{' regexpr '}' {repeatsymbol}
		| atom {repeatsymbol}
		;

repeatsymbol	: '*'
		| '+'
		;

 * nearatom takes care of range operator.  The nasty left recursion
 * has also been removed from this rule
 *
atomlist	: (nearatom)*
		;

nearatom	: anychar {'-' anychar}
		;

atom		: anychar

*/


#include <stdio.h>
#include "dlg.h"
#include "tokens.h"

int	action_no = 0;	/* keep track of actions outputed */
set	temp_label;	/* used to sum up possible labels on arc */
set	valid_chars;	/* mask to get rid elements that aren't used
			   in set */
int	flag_paren = FALSE;
int	flag_brace = FALSE;


grammar()
{
	used_chars = set_of(nil);
	/* optional actions section changed to allow multiple actions */
	while(!action())
		/* empty body */;
	/* THIS really SHOULDN'T be here */
	print_head();
	if (token != PER_PER)
		error("\"\%\%\" missing ",line_no);
	advance();
	if (rulelist())
		return TRUE;
	if (token != PER_PER)
		error("\"\%\%\" missing ",line_no);
	advance();
	/* do class compression here */
	if (compress)
		relabel();
	else{
		used_classes = used_chars;
		class_no = LAST_CHAR+2;
		}
	/* convert nfa to dfa */
	nfa_to_dfa(TOS.l);
	/* print out dfa in correct format */
	print_tables();
	/* optional actions at end changed to allow multiple actions */
	while(!action())
		/* empty body */;
	return FALSE;
}

rulelist()
{
	int	t;

	if (rule())
		return TRUE;
	while (!rule()){
		/* not really paths on stack any more, just index
		   to the regular expressions */
		t = NEW_NFA();
		nfa(t).trans[0] = NOS.l;
		nfa(t).trans[1] = TOS.l;
		NOS.l = t; NOS.r = NIL_INDEX;
		POP_PATH();
		}
	return FALSE;
}

rule()
{
	if (regexpr()){
		path	t;

		if (token == LESS_LESS){
			/* put something on the stack just as a place holder */
			t.l = NULL; t.r = NULL;
			PUSH_PATH(t);
		
			error("no expression for action  ",line_no);
			/* don't return TRUE; */
			}
		else
			return TRUE;
		}
	fprintf(OUT,"static\nact%d()\n{ ",++action_no);
	if (action())
		error("action missing ",line_no);
	fprintf(OUT,"}\n\n");
	/* mark accept state */
	nfa(TOS.r).accept = action_no;
	return FALSE;
}

regexpr(){
	/* check for unmatched } or ) at end of regular expression and
	   possibly more regexpr's */
	int failed;

	flag_paren = FALSE;
	flag_brace = FALSE;
	failed = regexpr1();
	/* time to see if there are surplus ')' or '}' */
	if (failed){
		while ((token == CLOSE_PAR) || (token == CLOSE_BRACE)){
			if ((!flag_paren) && (token == CLOSE_PAR)){
				error("unmatched \")\"  ",line_no);
				flag_paren = TRUE;
				}
			if ((!flag_brace) && (token == CLOSE_BRACE)){
				error("unmatched \"}\"  ",line_no);
				flag_brace = TRUE;
				}
			advance();
			}
	/* clean up other fragments of regular expression by resyncing
		on "<<" then */
		if (flag_paren || flag_brace)
			while(token != LESS_LESS)
				advance();
	}
	return failed;
}

regexpr1()
{
	int	t1,t2;

	if (andexpr())
		return TRUE;
	while (TRUE){
		if ( token != OR)
			break;
		advance();
		if (andexpr())
			return TRUE;
		t1 = NEW_NFA(); t2 = NEW_NFA();
		nfa(t1).trans[0] = NOS.l;
		nfa(NOS.r).trans[1] = t2;
		nfa(t1).trans[1] = TOS.l;
		nfa(TOS.r).trans[1] = t2;
		NOS.l = t1; NOS.r = t2;
		POP_PATH();
		}
	return FALSE;
}

andexpr()
{
	if (expr())
		return TRUE;
	while(!expr()){
		nfa(NOS.r).trans[1] = TOS.l;
		NOS.r = TOS.r;
		POP_PATH();
		}
	return FALSE;
}

expr()
{
	int complement = FALSE;

	switch(token){
	case NOT:	advance();
			complement = TRUE;
			if (token != OPEN_BRACK){
				error("no set to complement  ",line_no);
				/* ignore ~ and just treat as default */
				goto ignore_not;
				}
	case OPEN_BRACK: advance();
			if (atomlist(complement))
				return TRUE;
			if (token != CLOSE_BRACK)
				error("closing bracket assumed  ",line_no);
				/* don't return TRUE; */
			else
				advance();
			repeatsymbol();
			return FALSE;
	case OPEN_PAR:	advance();
			if (regexpr())
				return TRUE;
			if (token != CLOSE_PAR)
				error("closing parenthesis assumed  ",line_no);
				/* don't return TRUE; */
			else
				advance();
			repeatsymbol();
			return FALSE;
	case OPEN_BRACE: advance();
			if (regexpr())
				return TRUE;
			if (token != CLOSE_BRACE)
				error("closing brace assumed  ",line_no);
				/* don't return TRUE; */
			else
				advance();
			{ /* make optional */
				int t1, t2;

				t1 = NEW_NFA(); t2 = NEW_NFA();
				nfa(t1).trans[0] = TOS.l;
				nfa(t1).trans[1] = t2;
				nfa(TOS.r).trans[1] = t2;
				TOS.l = t1; TOS.r = t2;
			}
			repeatsymbol();
			return FALSE;
	ignore_not:
	default:	if (atom())
				return TRUE;
			/* add to used char */
			used_chars = set_or(used_chars,temp_label);
			repeatsymbol();
			return FALSE;
	}
}

repeatsymbol()
{
	int	t1,t2;

	switch(token){
	case ZERO_MORE:	advance();
			/* add epsilon arc to make repeatable */
			nfa(TOS.r).trans[0] = TOS.l;
			t1 = NEW_NFA(); t2 = NEW_NFA();
			nfa(t1).trans[0] = TOS.l;
			nfa(t1).trans[1] = t2;
			nfa(TOS.r).trans[1] = t2;
			TOS.l = t1; TOS.r = t2;
			return FALSE;
	case ONE_MORE:	advance();
			/* add epsilon arc to make repeatable */
			nfa(TOS.r).trans[0] = TOS.l;
			/* label should still be epsilon */
			return FALSE;
	default:	return TRUE;
	}
}

/* much simplified from the original grammar spec */
/* RANGE ('-') operator handled in nearatom */
atomlist(complement)
int complement;
{
	path	t;

	temp_label = set_of(nil);
	t.l = NEW_NFA();
	t.r = NEW_NFA();
	nfa(t.l).trans[0] = t.r;
	PUSH_PATH(t);
	while(!nearatom())
	/* labels accumlated in temp_label by nearatom() */;
	if (complement) 
		/* EOF will not be in the complement */
		temp_label = set_and(set_not(temp_label),
			valid_chars);
	/* add to used char */
	used_chars = set_or(used_chars,temp_label);
	nfa(t.l).label = temp_label;
	return FALSE;
}

action()
{

	if (token != LESS_LESS)
		return TRUE;
	scarf_action();
	if (token != GREAT_GREAT)
		return TRUE;
	advance();
	return FALSE;
}

nearatom()
{
	register int	b,e;

	if (token > MAXCHAR)
		return TRUE;
	/* all characters shift up one to make room for LEX_EOF */
	b = token+1;
	temp_label = set_or(temp_label, set_of(b));
	/* munch the character */
	advance();
	/* takes care of range operator, '-' */
	if (token != RANGE)
		return FALSE;
	advance();
	if (token > MAXCHAR){
		error("missing end of range",line_no);
		return TRUE;
		}
	/* all characters shift up one to make room for LEX_EOF */
	e = token+1;
	/* now fill in space between char b and e */
	while (b<=e)
		temp_label = set_or(temp_label,set_of(b++));
	/* munch the character */
	advance();
	return FALSE;
}

atom()
{
	path	t;

	if (token > MAXCHAR)
		return TRUE;
	/* build section of nfa */
	t.l = NEW_NFA();
	t.r = NEW_NFA();
	nfa(t.l).trans[0] = t.r;
	/* all characters shift up one to make room for LEX_EOF */
	nfa(t.l).label = set_of(token+1);
	PUSH_PATH(t);
	used_chars = set_or(used_chars, nfa(t.l).label);
	/* munch the character */
	advance();
	return FALSE;
}

int NEW_NFA()
{
	nfa_array[nfa_pool] = clear_nfa_node;
	return nfa_pool++;
}
